import java.io.*;
import java.util.Random;

public final class Codegen {
	public static String fileNameC = null;
	public static String fileNameH = null;
	
	public static int Ncut;
	public static int DIM;
	public static int LENGTH;

	public static final void main(String[] Args)
	{
		if (Args.length > 0) {
			try {
				Ncut = Integer.parseInt(Args[0]);
				DIM = Integer.parseInt(Args[1]);
				LENGTH = Integer.parseInt(Args[2]);
			} catch (NumberFormatException e) {
				System.err.println("Argument must be an integer");
				System.exit(1);
			}
		}

		generate_global_h();
		//Source Code Gen
		int i, j, k, l;
		//i = j
		for(i = 0; i < Ncut; i++)
		{
			processGenii(i);
			headerGenii(i);
		}
		//i != j
		for(i = 0; i < Ncut; i++)
		{
			for(j = i+1; j < Ncut; j++)
			{
				processGenij(j, i);
				headerGenij(j, i);
			}
		}
	}
	public static void generate_global_h()
	{
		try {
			BufferedWriter outG = new BufferedWriter(new FileWriter("global.h", true));
			outG.write("#ifndef GLOBAL_H\n");
			outG.write("#define GLOBAL_H\n\n");
	
			outG.write("#define DIM "+DIM+"\n");	//Dimension
			outG.write("#define NCUT "+Ncut+"\n");	//Number of cuts
			outG.write("#define N DIM/NCUT\n");	//N*N dimension sub-matrices	
			outG.write("#define LENGTH "+LENGTH+"\n");	//Number of generated matrices
	
			outG.write("typedef double matrix[N][N];\n");
	
			outG.write("#endif\n");
			outG.close();
		}
		catch (IOException e) {}
	}
	public static void processInitiate(BufferedWriter out, int i, int j)
	{
		
		try {
			out.write("#include <stdio.h>\n");
			out.write("#include <math.h>\n");
			out.write("#include \"p"+"_"+i+"_"+j+".h\"\n");
			if(i == j && i == 0)
				out.write("#include \"matrix.c\"\n");
			out.write("\n");
			if(i == j)
			{
				if(i > 0)
				{
					out.write("void TMultiply(matrix A, matrix C);\n");
					out.write("void Subtract(matrix A, matrix B);\n");
				}
				out.write("void Cholesky(matrix A, matrix L);\n");
				if(i < Ncut-1)
				{
					out.write("void Invert(matrix x, matrix L);\n");
					out.write("void Transpose(matrix A, matrix B);\n");
				}
			}
			else
			{
				out.write("void Multiply(matrix A, matrix B, matrix C);\n");
				if(j > 0)
					out.write("void Subtract(matrix A, matrix B);\n");
				if(i < Ncut - 1)
					out.write("void Transpose(matrix A, matrix B);\n");
			}	
			out.write("\n");
			out.write("void p_"+i+"_"+j+"_init(DOLProcess *p)\n");
			out.write("{\n");
			out.write("\tp->local->index = 0;\n");
			out.write("\tp->local->len = LENGTH;\n");
			out.write("}\n");

			out.write("int p_"+i+"_"+j+"_fire(DOLProcess *p)\n"); 
			out.write("{");
    			out.write("\tif (p->local->index < p->local->len)\n"); 
			out.write("\t{\n");
			out.write("\t\tDOL_read((void*)PORT_IN_GEN, p->local->a, N*N*sizeof(double), p);\n\n");
		}
		catch (IOException e) {}
	}
	
	public static void processGenii(int i)
	{
		int j, k;
		processNameGen(i, i);
		try {
			BufferedWriter outC = new BufferedWriter(new FileWriter(fileNameC, true));
			
			processInitiate(outC, i, i);

			for(k = 0; k < i; k++)
			{
				outC.write("\t\tDOL_read((void*)PORT_IN_"+i+"_"+k+", p->local->b, N*N*sizeof(double), p);\n");
				//Transpose & Multiply
				outC.write("\t\tTMultiply(p->local->b, p->local->c);\n");
				outC.write("\t\tSubtract(p->local->a, p->local->c);\n\n");
			}   	
			outC.write("\t\tCholesky(p->local->a, p->local->L);\n");
			if(i < Ncut - 1)
			{
				outC.write("\t\tInvert(p->local->x, p->local->L);\n");
				outC.write("\t\tTranspose(p->local->x, p->local->Inv);\n\n");
			}
			outC.write("\t\tDOL_write((void*)PORT_OUT_CON, p->local->L, N*N*sizeof(double), p);\n");
			for(j = i+1; j < Ncut; j++)
			{
				outC.write("\t\tDOL_write((void*)PORT_OUT_"+j+"_"+i+", p->local->Inv, N*N*sizeof(double), p);\n");
			}
			processFinalize(outC);
			outC.close();
		}
		catch (IOException e) {}
	}

	public static void headerGenii(int i)
     {
		int j, k;
		headerNameGen(i, i);
		try {
			BufferedWriter outH = new BufferedWriter(new FileWriter(fileNameH, true));
			
			outH.write("#ifndef P"+"_"+i+"_"+i+"_H\n");
			outH.write("#define P"+"_"+i+"_"+i+"_H\n\n");
			outH.write("#include <dol.h>\n");
			outH.write("#include \"global.h\"\n\n");
			
			//Header define ports
			outH.write("#define PORT_IN_GEN \"port_generator\"\n");
			for(k = 0; k < i; k++)
			{
				outH.write("#define PORT_IN_"+i+"_"+k+" \"IN"+"x"+i+"x"+k+"\"\n");
			}
               for(j = i+1; j < Ncut; j++)
			{
				outH.write("#define PORT_OUT_"+j+"_"+i+" \"OUT"+"x"+j+"x"+i+"\"\n");
			}
			outH.write("#define PORT_OUT_CON \"port_consumer\"\n\n");
			outH.write("typedef struct _local_states {\n");
			outH.write("\tint index;\n");
			outH.write("\tint len;\n");
			
			outH.write("\tmatrix a;\n");
			outH.write("\tmatrix L;\n");
			if(i != 0)
			{
				outH.write("\tmatrix b;\n");
				outH.write("\tmatrix c;\n");
			}	
			if(i < Ncut - 1)
			{
				outH.write("\tmatrix x;\n");
				outH.write("\tmatrix Inv;\n");
			}
			outH.write("} P"+"_"+i+"_"+i+"_State;\n\n");
			outH.write("void p"+"_"+i+"_"+i+"_init(DOLProcess *);\n");
			outH.write("int p"+"_"+i+"_"+i+"_fire(DOLProcess *);\n\n");
			outH.write("#endif\n");
	
			outH.close();
		}
		catch (IOException e) {}
	}
	
	public static void processGenij(int i, int j)
	{
		int x;
		try {
			processNameGen(i, j);
			BufferedWriter outC = new BufferedWriter(new FileWriter(fileNameC, true));
			processInitiate(outC, i, j);
			for(x = 0; x < j; x++)
			{
				outC.write("\t\tDOL_read((void*)PORT_IN_"+i+"_"+x+", p->local->pH, N*N*sizeof(double), p);\n");
				outC.write("\t\tDOL_read((void*)PORT_IN_"+j+"_"+x+", p->local->pD, N*N*sizeof(double), p);\n");
				outC.write("\t\tMultiply(p->local->pH, p->local->pD, p->local->pp);\n");
				outC.write("\t\tSubtract(p->local->a, p->local->pp);\n\n");
			}
			outC.write("\t\tDOL_read((void*)PORT_IN_"+j+"_"+j+", p->local->Inv, N*N*sizeof(double), p);\n");
			outC.write("\t\tMultiply(p->local->a, p->local->Inv, p->local->C);\n\n");
   	 		outC.write("\t\tDOL_write((void*)PORT_OUT_CON, p->local->C, N*N*sizeof(double), p);\n");
			for(x = j+1; x < i+1; x++)	//OUT horizontal
			{
				outC.write("\t\tDOL_write((void*)PORT_OUT_"+i+"_"+x+", p->local->C, N*N*sizeof(double), p);\n");
			}
			for(x = i+1; x < Ncut; x++)	//OUT Diagonal
			{   
				if(x == i+1)
					outC.write("\n\t\tTranspose(p->local->C, p->local->CT);\n");
				outC.write("\t\tDOL_write((void*)PORT_OUT_"+x+"_"+i+", p->local->CT, N*N*sizeof(double), p);\n");
			}
			processFinalize(outC);
		   	outC.close();
		}
		catch (IOException e) {}	
	}
	
	public static void headerGenij(int i, int j)
	{
		int x;
		headerNameGen(i, j);
		try {
			BufferedWriter outH = new BufferedWriter(new FileWriter(fileNameH, true));

			outH.write("#ifndef P"+"_"+i+"_"+j+"_H\n");
			outH.write("#define P"+"_"+i+"_"+j+"_H\n\n");
			outH.write("#include <dol.h>\n");
			outH.write("#include \"global.h\"\n\n");

			//Header define ports
			outH.write("#define PORT_IN_GEN \"port_generator\"\n");
			for(x = 0; x < j; x++)
			{
				outH.write("#define PORT_IN_"+i+"_"+x+" \"IN"+"x"+i+"x"+x+"\"\n");
				outH.write("#define PORT_IN_"+j+"_"+x+" \"IN"+"x"+j+"x"+x+"\"\n");
			}
			outH.write("#define PORT_IN_"+j+"_"+j+" \"IN"+"x"+j+"x"+j+"\"\n");
			for(x = j+1; x < i+1; x++)      //OUT horizontal
			{
				outH.write("#define PORT_OUT_"+i+"_"+x+" \"OUT"+"x"+i+"x"+x+"\"\n");
			}
			for(x = i+1; x < Ncut; x++)     //OUT Diagonal
			{
				outH.write("#define PORT_OUT_"+x+"_"+i+" \"OUT"+"x"+x+"x"+i+"\"\n");
			}
			outH.write("#define PORT_OUT_CON \"port_consumer\"\n\n");
			outH.write("typedef struct _local_states {\n");
			outH.write("\tint index;\n");
			outH.write("\tint len;\n");

			outH.write("\tmatrix a;\n");
			outH.write("\tmatrix Inv;\n");
			outH.write("\tmatrix C;\n");
			if(j != 0)
			{
				outH.write("\tmatrix pH;\n");
				outH.write("\tmatrix pD;\n");
				outH.write("\tmatrix pp;\n");	
			}
			if(i < Ncut - 1)
			{
				outH.write("\tmatrix CT;\n");
			}
			outH.write("} P"+"_"+i+"_"+j+"_State;\n\n");
			outH.write("void p"+"_"+i+"_"+j+"_init(DOLProcess *);\n");
			outH.write("int p"+"_"+i+"_"+j+"_fire(DOLProcess *);\n\n");
			outH.write("#endif\n");

			outH.close();
		}
		catch (IOException e) {}
	}
	
	public static void processFinalize(BufferedWriter out)
	{
		try {
		out.write("\t\tp->local->index++;\n");
		out.write("\t}\n\n");
		
		out.write("\tif (p->local->index >= p->local->len)\n");
		out.write("\t{\n");
		out.write("\t\tDOL_detach(p);\n");
		out.write("\t\treturn -1;\n");
		out.write("\t}\n");
		out.write("\treturn 0;\n");
		out.write("}\n");
		}
		catch (IOException e) {}
	}

	public static void processNameGen(int i, int j)
	{
		fileNameC = "p" + "_" + i + "_" + j + ".c";
	}
	
	public static void headerNameGen(int i, int j)
	{
		fileNameH = "p" + "_" + i + "_" + j + ".h";
	}

	public static void Send(String matrix, int x, BufferedWriter outC, BufferedWriter outH)
	{
		try
		{
			String port = "OUT" + x;
			outC.write("\nDOL_write((void*)"+port+", p->local->"+matrix+");");
			outH.write("\n#define "+port+" "+x);
		}
		catch (IOException e) {}
	}
	public static void Initiate(int k, BufferedWriter outC, BufferedWriter outH)
	{
		try
		{
			outC.write("\nNow in a function\n");
		}
		catch(IOException e) {}
	}													
}
 

